(* Code generation for Go-- *)

module L = Llvm
module A = Ast
open Sast 

module StringMap = Map.Make(String)

(* translate : Sast.program -> Llvm.module *)
let translate (sdefs,(globals, functions)) =
  let context    = L.global_context () in
  
  (* Create the LLVM compilation module into which
     we will generate code *)
  let the_module = L.create_module context "Gmm" in

  (* Get types from the context *)
  let i32_t      = L.i32_type    context
  and i64_t      = L.i64_type    context
  and i8_t       = L.i8_type     context
  and i1_t       = L.i1_type     context
  and float_t    = L.double_type context
  and ptr_t      = L.pointer_type (L.i8_type context)
  and void_t     = L.void_type   context in
  let mutex_t    = L.struct_type context [|i64_t; L.array_type i8_t 56|]
  and sem_t      = L.struct_type context [| i64_t; L.array_type i8_t 24|] in
  
  (* collect struct definition in a StringMap hash table in Ocaml *)
  (* Map structure { key:sID ,value: (idx in the struct, (type,name)) } *)
  let rec seq n = if n >= 0 then (n :: (seq (n-1))) else [] in
  let sdecl_map = 
	let process_def m sdef = (match sdef with
		A.StructDef(id,binds) ->
		let len = List.length binds in
		let seqlist = List.rev (seq (len-1)) in
		let values = List.combine seqlist binds in
		StringMap.add id values m)
	in
  	List.fold_left process_def StringMap.empty sdefs  
  in		                         
  (* Return the LLVM type for a MicroC type *)
  let rec ltype_of_typ = function
      A.Int   -> i32_t
    | A.Bool  -> i1_t
    | A.Float -> float_t
    | A.Void  -> void_t
    | A.Array(ty) -> L.pointer_type (ltype_of_typ ty)
    | A.String -> ptr_t
    | A.Channel(ty) -> 
      let ptrs = List.map L.pointer_type [ltype_of_typ ty; i32_t; mutex_t; sem_t] in
      let ptrs_arr = Array.of_list ptrs in
      let chan_t = L.struct_type context ptrs_arr in
      let chanptr = L.pointer_type chan_t in
      chanptr 
    | A.Struct(s) ->
      let mvalues = StringMap.find s sdecl_map in
      let collect_typ (_,(t,_)) = ltype_of_typ t in
      let typs = List.map collect_typ mvalues in
      let typs_arr = Array.of_list typs in
      let struct_t = L.struct_type context typs_arr in
      let struct_ptr = L.pointer_type struct_t in
      struct_ptr
    in

  (* Create a map of global variables after creating each *)
  let global_vars : L.llvalue StringMap.t =
    let global_var m (t, n) = 
      let init = match t with
          A.Float -> L.const_float (ltype_of_typ t) 0.0
        | A.Array(_) | A.Channel (_) | A.String |A.Struct(_) -> L.const_null (ltype_of_typ t)
        | _ -> L.const_int (ltype_of_typ t) 0
      in StringMap.add n (L.define_global n init the_module) m in
    List.fold_left global_var StringMap.empty globals in
    
  (* built-in functions *)
  let printf_t : L.lltype = 
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func : L.llvalue = 
      L.declare_function "printf" printf_t the_module in
  (* semaphore calls for channel *)
  let sem_init_t : L.lltype = L.function_type i32_t [|L.pointer_type sem_t; i32_t; i32_t|] in
  let sem_init_f : L.llvalue = L.declare_function "sem_init" sem_init_t the_module in
  (* functions for strings *)
  let strcat_t : L.lltype = 
	  L.function_type ptr_t [|ptr_t;ptr_t|] in
  let strcat_f : L.llvalue =
    L.declare_function "string_cat" strcat_t the_module in
  let strcmp_t : L.lltype =
      L.function_type i32_t [|ptr_t;ptr_t|] in
  let strcmp_f : L.llvalue =
    L.declare_function "string_cmp" strcmp_t the_module in
 
  (* Define each function (arguments and return type) so we can 
     call it even before we've created its body *)
  let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types = 
	      Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
      in 
      let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
    StringMap.add name (L.define_function name ftype the_module, fdecl) m in
  List.fold_left function_decl StringMap.empty functions in
  
  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder
    and str_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

    (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
    let local_vars =
      let add_formal m (t, n) p =
        L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ t) n builder in
        ignore (L.build_store p local builder);
        StringMap.add n local m 
      and add_local m (t, n) =
        let local_var = match t with 
           _ -> L.build_alloca (ltype_of_typ t) n builder
        in StringMap.add n local_var m
      in
      let formals = List.fold_left2 add_formal StringMap.empty fdecl.sformals
          (Array.to_list (L.params the_function)) in
      List.fold_left add_local formals fdecl.slocals 
    in

    (* Return the value for a variable or formal argument.
       Check local names first, then global names *)
    let lookup n = try StringMap.find n local_vars
                   with Not_found -> StringMap.find n global_vars
    in

    (* some util functions *)
    let construct_chan_t ty = 
      let ptrs = List.map L.pointer_type [ltype_of_typ ty; i32_t; mutex_t; sem_t] in
      let ptrs_arr = Array.of_list ptrs in
      let chan_t = L.struct_type context ptrs_arr in
      chan_t
    in

    let lnum i = L.const_int i32_t i in
    let insert_value builder agg i v = 
      L.build_insertvalue agg v i "tmp__" builder 
    in

    let inc_seq k = 
      let rec dec_seq n = 
        if n >= 0 then (n :: (dec_seq (n-1))) 
        else []
      in List.rev (dec_seq k)
    in
    
    (* Construct code for an expression; return its value *)
    let rec expr builder ((_, e) : sexpr) = match e with
	      SLiteral i  -> L.const_int i32_t i
      | SStrLit s   -> L.build_global_stringptr s "str" builder
      | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
      | SFliteral l -> L.const_float_of_string float_t l
      | SNoexpr     -> L.const_int i32_t 0
      | SId s       -> L.build_load (lookup s) s builder
      (* array element assignment *)
      | SAssign((_, SArrAccess(e1, e2)), e3) ->
        let new_v = expr builder e3 in
        let arr = expr builder e1 in 
        let idx = expr builder e2 in
        let ptr = L.build_gep arr [|idx|] "" builder in
        ignore(L.build_store new_v ptr builder); new_v
        (* struct element assignment *)
      | SAssign((_, SDot((A.Struct(sname),e1),fname)),e2) ->
        let new_v = expr builder e2 in
		    let stptr = expr builder (A.Struct(sname),e1) in
        let st = L.build_load stptr "structval" builder in
        let binds = StringMap.find sname sdecl_map in
		    let rec idx l name = match l with [] -> raise (Failure "field not found")
                           | (i,(_,n))::tl -> if n = name then i else idx tl name in
        let index = idx binds fname in
        let new_struct = L.build_insertvalue st new_v index "dot" builder in
        ignore(L.build_store new_struct stptr builder); new_v
      | SAssign (e1, e2) ->
        let e' = expr builder e2 in
        (match snd e1 with 
           SId s -> ignore(L.build_store e' (lookup s) builder); e'
          | _ -> raise (Failure ("assignment for " ^ (string_of_sexpr e2) ^ 
                                  "not implemented in codegen")))
	    (* intializing a new struct*)
      | SNStruct(ty,el) ->
        let fields = List.map (expr builder) el in
        let tuples = match ty with A.Struct(s) -> StringMap.find s sdecl_map | _ ->raise (Failure "not struct type") in
        let types = List.map (fun (_,(t,_)) -> ltype_of_typ t) tuples in
        let length = List.length types in
        let idxs = List.rev (seq (length-1)) in
        let llstruct_t = L.struct_type context (Array.of_list types) in
        let ptr = L.build_malloc llstruct_t "struct_init" builder in
        let value = List.fold_left2 (insert_value builder) (L.const_null llstruct_t) idxs fields in
        let _ = L.build_store value ptr builder in ptr
      | SDot((A.Struct(sname),e),fname) ->
        let stptr = expr builder (A.Struct(sname),e) in
        let binds = StringMap.find sname sdecl_map in
        let rec idx l name = match l with [] -> raise (Failure "field not found")
                    | (i,(_,n))::tl -> if n = name then i else idx tl name in 
        let index = idx binds fname in
        let svalue = L.build_load stptr "structval" builder in
        L.build_extractvalue svalue index fname builder
      (* array implementation on stack, doesn't need to be on the heap for now*)
      | SNArray(ty, e) -> 
        let arr_size = expr builder e in
        let llarr_typ = ltype_of_typ ty in
        let ptr = L.build_array_alloca llarr_typ arr_size "" builder in ptr
      (* array access *)
      | SArrAccess(e1, e2) -> 
        let arr = expr builder e1 in
        let idx = expr builder e2 in
        let ptr = L.build_load (L.build_gep arr [|idx|] "" builder) "" builder in
        ptr
       (* channel initialization *)
      | SNChan(ty, e) ->
        let malloc_chan lsize ty =
          let typs = [ltype_of_typ ty; i32_t; mutex_t; sem_t] in
          let nums = [lsize; lnum 4; lnum 1; lnum 2] in
          let my_arr_malloc a b = L.build_array_malloc a b "arr_malloc_" builder in
          let ptrs = List.map2 my_arr_malloc typs nums in
          (* assign values to the book-keeping array *)
          let my_store agg idx value= 
            let store_ptr = L.build_gep agg [|idx|] "arr_gep_" builder in
            ignore(L.build_store value store_ptr builder); store_ptr
          in 
          let type_id = match ty with
              A.Int -> lnum 0
            | A.Float -> lnum 1
            | A.Bool -> lnum 2
            | A.String -> lnum 3
            | A.Struct(_) -> lnum 4
            | _ -> raise (Failure "type is not supported by channel")
          in
          let bkptr = List.nth ptrs 1 in
          let _ = List.map2 (my_store bkptr) [(lnum 0); (lnum 3)] [lsize; type_id] in
          (* sem_init on the two semaphore *)
          let sem_arr = List.nth ptrs 3 in
          let sem_empty = L.build_gep sem_arr [|lnum 0|] "sem_access_" builder in 
          let _ = L.build_call sem_init_f [|sem_empty; (lnum 0); (lnum 0)|] "sem_empty" builder in
          let sem_full = L.build_gep sem_arr [|lnum 1|] "sem_access_" builder in
          let _ = L.build_call sem_init_f [|sem_full; (lnum 0); lsize |] "sem_empty" builder in
          ptrs
        in
        let chan_size = expr builder e in
        let chan_t = construct_chan_t ty in
        let vals = malloc_chan chan_size ty in
        let seq = inc_seq ((List.length vals)-1) in
        let llvals = List.fold_left2 
         (fun agg i v -> insert_value builder agg i v)
         (L.const_null chan_t) seq vals in
        let chanptr = L.build_malloc chan_t "chanInit" builder in
        ignore(L.build_store llvals chanptr builder);chanptr
        (* channel enqueue/dequeue op *)
       | SEnqueue(e1, e2) -> 
        let new_v = expr builder e2 in
        let chanptr = expr builder e1 in
        let chan_v = L.build_load chanptr "chan_val" builder in
        let (tyy, _ ) = e2 in
        (* get pointer to value to be enqueued --> copy by value *)
        let val_ptr = L.build_malloc (ltype_of_typ tyy) "val_ptr_" builder in
        let _ = L.build_store new_v val_ptr builder in
        (* construct the function call *)
        let enc_t : L.lltype =
          let lltype = ltype_of_typ tyy in
          let tylist = [lltype; i32_t; mutex_t; sem_t; lltype] in
          let typtr = List.map L.pointer_type tylist in
          let tyarr = Array.of_list typtr in
          L.function_type i32_t tyarr in
        let enc_f : L.llvalue = L.declare_function "enqueue_f" enc_t the_module in
        (* extract value *)
        let my_extract agg idx = L.build_extractvalue agg idx "extrac_" builder in
        let vals = List.map (my_extract chan_v) (inc_seq 3) in
        let _ = L.build_call enc_f (Array.of_list (List.append vals [val_ptr])) "enc_ff" builder in
        new_v
      | SDequeue(e1, e2) ->
        let chanptr = expr builder e1 in
        let chan_v = L.build_load chanptr "chan_val" builder in
        let tyy =  match fst e1 with
            A.Channel(tyy) -> tyy
          | _ -> raise (Failure "dequeue from incorrect type")
        in
        let val_ptr = L.build_alloca (ltype_of_typ tyy) "val_ptr_" builder in
        (* construct the function call *)
        let dec_t : L.lltype =
          let lltype = ltype_of_typ tyy in
          let tylist = [lltype; i32_t; mutex_t; sem_t; lltype] in
          let typtr = List.map L.pointer_type tylist in
          let tyarr = Array.of_list typtr in
          L.function_type i32_t tyarr in
        let dec_f : L.llvalue = L.declare_function "dequeue_f" dec_t the_module in
        (* extract value *)
        let my_extract agg idx = L.build_extractvalue agg idx "extrac_" builder in
        let vals = List.map (my_extract chan_v) (inc_seq 3) in
        let _ = L.build_call dec_f (Array.of_list (List.append vals [val_ptr])) "enc_ff" builder in
        let new_v = L.build_load val_ptr "new_v_dec_" builder in
        let eval = match e2 with
          (A.Void, _) -> new_v
          |(_, SId s) -> ignore(L.build_store new_v (lookup s) builder); new_v
          | _ -> raise (Failure ("assignment for " ^ (string_of_sexpr e2) ^ 
                                  "not implemented in codegen"))
        in eval
      | SBinop ((A.Float,_ ) as e1, op, e2) ->
        let e1' = expr builder e1
        and e2' = expr builder e2 in
        (match op with 
          A.Add     -> L.build_fadd
        | A.Sub     -> L.build_fsub
        | A.Mult    -> L.build_fmul
        | A.Div     -> L.build_fdiv 
        | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
        | A.Neq     -> L.build_fcmp L.Fcmp.One
        | A.Less    -> L.build_fcmp L.Fcmp.Olt
        | A.Leq     -> L.build_fcmp L.Fcmp.Ole
        | A.Greater -> L.build_fcmp L.Fcmp.Ogt
        | A.Geq     -> L.build_fcmp L.Fcmp.Oge
        | A.And | A.Or ->
          raise (Failure "internal error: semant should have rejected and/or on float")
        ) e1' e2' "tmp" builder
      | SBinop ((A.String,_ ) as e1, op, e2) ->
        let e1' = expr builder e1
        and e2' = expr builder e2 in
        (match op with 
          A.Add     -> L.build_call strcat_f [| e1';e2' |] "string_cat" builder
        | A.Equal   -> (L.build_icmp L.Icmp.Eq) (L.const_int i32_t 0) 
					   (L.build_call strcmp_f [| e1';e2' |] "string_cmp" builder) "tmp" builder
        | A.Neq     -> (L.build_icmp L.Icmp.Ne) (L.const_int i32_t 0)
					   (L.build_call strcmp_f [| e1';e2' |] "string_cmp" builder) "tmp" builder
        | _		    ->
          raise (Failure "internal error: semant should have rejected and/or on float")
        )       
      | SBinop (e1, op, e2) ->
        let e1' = expr builder e1
        and e2' = expr builder e2 in
        (match op with
          A.Add     -> L.build_add
        | A.Sub     -> L.build_sub
        | A.Mult    -> L.build_mul
        | A.Div     -> L.build_sdiv
        | A.And     -> L.build_and
        | A.Or      -> L.build_or
        | A.Equal   -> L.build_icmp L.Icmp.Eq
        | A.Neq     -> L.build_icmp L.Icmp.Ne
        | A.Less    -> L.build_icmp L.Icmp.Slt
        | A.Leq     -> L.build_icmp L.Icmp.Sle
        | A.Greater -> L.build_icmp L.Icmp.Sgt
        | A.Geq     -> L.build_icmp L.Icmp.Sge
        ) e1' e2' "tmp" builder
      | SUnop(op, ((t, _) as e)) ->
        let e' = expr builder e in
        (match op with
            A.Neg when t = A.Float -> L.build_fneg 
          | A.Neg                  -> L.build_neg
          | A.Not                  -> L.build_not
        ) e' "tmp" builder
      | SFuncCall ("print", [e]) | SFuncCall ("printb", [e]) ->
        L.build_call printf_func [| int_format_str ; (expr builder e) |]
        "printf" builder
      | SFuncCall ("printf", [e]) -> 
	      L.build_call printf_func [| float_format_str ; (expr builder e) |]
	      "printf" builder
      | SFuncCall ("prints",[e]) ->
        L.build_call printf_func [| str_format_str ; (expr builder e) |]
	      "printf" builder
      | SFuncCall (f, args)  ->
        let (fdef, fdecl) = StringMap.find f function_decls in
        let llargs = List.rev (List.map (expr builder) (List.rev args)) in
        let result = (match fdecl.styp with 
                          A.Void -> ""
                        | _ -> f ^ "_result") in
          L.build_call fdef (Array.of_list llargs) result builder
      | SGofuncCall(f, args) ->
        (* TODO: Now Gocall only supports up to 3 float/int arguments*)
        let (fdef, fdecl) = StringMap.find f function_decls in
        (* convert number to a desc. sequence *)
        let rec seq n = if n >= 0 then (n :: (seq (n-1))) else [] in
        (* match primitive types to unique number to allow external C functions
           to identify them.*)
        let match_type = function
             A.Int   -> 1
           | A.Float -> 2
           | _ 		 -> 3 	in (* pointer types: string, array, struct *)
        let offset ty = lnum (match_type ty) in
        (* get a list of length *)
        let typ_offsets alist (t,_) =((offset t)::alist) in
        let offsets = List.rev (List.fold_left typ_offsets [] fdecl.sformals) in
        (* collect types of each argument in the argument list *)
        let collect_typ alist (t,_) = ((ltype_of_typ t)::alist) in
        let typs = List.rev (List.fold_left collect_typ [] fdecl.sformals) in
        (* evaluate arguments and put them into (type,value) tuples *)
        let llargs = List.map (expr builder) args in
        let combine alist typ arg = ((typ,arg)::alist)in
        let tuples = List.rev (List.fold_left2 combine [] typs llargs) in
		    (* get the argument list length *)
        let arg_size l = List.fold_left (fun count _ -> count + 1) 0 l in
        let az = arg_size args in
        let count = L.const_int i32_t az in
        (* construct a struct called prefix that contains information
          * about gofunction arguments, the form of the prefix list is
          * [arg.length,arg1.type,arg2.type.....]
         *)
        let build_arg bty tylist len =
          let ty_idxs =  List.rev (seq ((List.length tylist) - 1)) in
          let ty_typs = Array.make len bty in
          let ty_typ = L.struct_type context ty_typs in
          let ty_ptr = L.build_alloca ty_typ "tmp_2" builder in
          let value = List.fold_left2 (insert_value builder) (L.const_null ty_typ) ty_idxs tylist in
          let _ = L.build_store value ty_ptr builder in
          (ty_typ, ty_ptr) in
        let prelist = (count::offsets) in
        let (pre_typ, pre) = build_arg i32_t prelist (az+1) in
        let build_nums filter bty = 
          let tylist = List.rev (List.fold_left filter [] tuples) in
          build_arg bty tylist (List.length tylist) in
        let collect_int alist (t,v) = (if t == i32_t then (v::alist) else alist) in
        let collect_double alist (t,v) = (if t == float_t then (v::alist) else alist) in
        let (int_typ, intptr) = build_nums collect_int i32_t
        and (d_typ, dptr) = build_nums collect_double float_t in 
        (* get a list of length *)
        let typ_matches alist (t,_) =((match_type t)::alist) in
        let matches = List.rev (List.fold_left typ_matches [] fdecl.sformals) in
        let tuples = List.rev (List.fold_left2 combine [] matches tuples) in
        let collect_ptrs tlist (o,(_,v)) = (if o = 3 then (v::tlist) else tlist) in
        let palist = List.rev (List.fold_left collect_ptrs [] tuples) in
        let cast some_p = L.build_bitcast some_p ptr_t "bitcast" builder in
        let palist = List.map cast palist in
        let (p_typ, pptr) = build_arg ptr_t palist (List.length palist) in
        let gocall_t : L.lltype =
            let func_ptr = L.pointer_type (L.function_type void_t (Array.of_list typs)) in
            (* external function takes pointers to 1.prefix list 2.int arg list 3. double arg list 4.struct/string arg list*)
        L.function_type i32_t [| func_ptr; (L.pointer_type pre_typ);(L.pointer_type int_typ);(L.pointer_type d_typ);(L.pointer_type p_typ) |] in
            let gocall_func : L.llvalue =
            L.declare_function "gocall" gocall_t the_module in
            L.build_call gocall_func [| fdef;pre;intptr;dptr;pptr |] "gocall" builder
	 | _ -> raise (Failure "expression not implemented") in

    
    (* LLVM insists each basic block end with exactly one "terminator" 
       instruction that transfers control.  This function runs "instr builder"
       if the current block does not already have a terminator.  Used,
       e.g., to handle the "fall off the end of the function" case. *)
    let add_terminal builder instr =
      match L.block_terminator (L.insertion_block builder) with
      Some _ -> ()
      | None -> ignore (instr builder) in
	
    (* Build the code for the given statement; return the builder for
       the statement's successor (i.e., the next instruction will be built
       after the one generated by this call) *)

    let rec stmt builder = function
	      SBlock sl -> List.fold_left stmt builder sl
      | SExpr e -> ignore(expr builder e); builder 
      | SReturn e -> ignore(match fdecl.styp with
                              (* Special "return nothing" instr *)
                              A.Void -> L.build_ret_void builder 
                              (* Build return statement *)
                            | _ -> L.build_ret (expr builder e) builder );
                     builder
      | SIf (predicate, then_stmt, else_stmt) ->
         let bool_val = expr builder predicate in
	       let merge_bb = L.append_block context "merge" the_function in
         let build_br_merge = L.build_br merge_bb in (* partial function *)
	       let then_bb = L.append_block context "then" the_function in
	           add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
	           build_br_merge;
	      let else_bb = L.append_block context "else" the_function in
	           add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
	           build_br_merge;
	           ignore(L.build_cond_br bool_val then_bb else_bb builder);
	           L.builder_at_end context merge_bb

      | SWhile (predicate, body) ->
        let pred_bb = L.append_block context "while" the_function in
        ignore(L.build_br pred_bb builder);

        let body_bb = L.append_block context "while_body" the_function in
        add_terminal (stmt (L.builder_at_end context body_bb) body)
          (L.build_br pred_bb);

        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = expr pred_builder predicate in

        let merge_bb = L.append_block context "merge" the_function in
        ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb

      (* Implement for loops as while loops *)
      | SFor (e1, e2, e3, body) -> stmt builder
	    ( SBlock [SExpr e1 ; SWhile (e2, SBlock [body ; SExpr e3]) ] )
    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (SBlock fdecl.sbody) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.styp with
        A.Void -> L.build_ret_void
      | A.Float -> L.build_ret (L.const_float float_t 0.0)
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
