(* Semantic checking for the Go-- compiler *)

open Ast
open Sast

module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (sdefs,(globals, functions)) =

  (* struct_defs*)
  let struct_defs = 
	let add_sdefs sds = 
  		let add_sdef m sd = match sd with StructDef(n,binds) ->
		  if StringMap.mem n m then raise (Failure ("duplicate structdef " ^ n)) else
			StringMap.add n binds m
		in List.fold_left add_sdef StringMap.empty sds
  	in add_sdefs sdefs
  in


  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    List.iter (function
	    (Void, b) | (Array(Void), b) | (Channel(Void), b) -> raise (Failure ("illegal void " ^ kind ^ " " ^ b))
      | _ -> ()) binds;
    let rec dups = function
        [] -> ()
      |	((_,n1) :: (_,n2) :: _) when n1 = n2 ->
	  raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in dups (List.sort (fun (_,a) (_,b) -> compare a b) binds)
  in

  (**** Check global variables ****)

  check_binds "global" globals;

  (**** Check functions ****)

  (* Collect function declarations for built-in functions: no bodies *)
  let built_in_decls = 
    let add_bind map (name, ty) = StringMap.add name {
      isGo = false; (* default to function for built-ins *)
      typ = Void; (* these built in functions return void *)
      fname = name; 
      formals = [(ty, "x")];
      locals = []; body = [] } map
    (* and add_bind2 map (name, ) *)
    in List.fold_left add_bind StringMap.empty [ ("print", Int);
			                         ("printb", Bool);
			                         ("printf", Float);
                               ("prints", String);
                               ]
  in

  (* Add function name to symbol table *)
  let add_func map fd = 
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
        (* duplicate of built-in funcs *)
         _ when StringMap.mem n built_in_decls -> make_err built_in_err
         (* duplicate of existing funcs *)
       | _ when StringMap.mem n map -> make_err dup_err  
       | _ ->  StringMap.add n fd map 
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_decls functions
  in
  
  (* Return a function from our symbol table *)
  let find_func s = 
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = 
    let main_func = find_func "main" in
    if main_func.isGo
    then raise (Failure "main has to be regular function")
  in  (* Ensure "main" is defined and is a regular function*)

  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    check_binds "formal" func.formals;
    check_binds "local" func.locals;

    (* Raise an exception if the given rvalue type cannot be assigned to
       the given lvalue type *)
    let check_assign lvaluet rvaluet err =
       if lvaluet = rvaluet then lvaluet else raise (Failure err)
    in   

    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
	                StringMap.empty (globals @ func.formals @ func.locals )
    in

    (* Return a variable from our local symbol table *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    (* Return a semantically-checked expression, i.e., with a type *)
    let rec expr = function
        Literal  l -> (Int, SLiteral l)
      | StrLit l -> (String,SStrLit l)
      | Fliteral l -> (Float, SFliteral l)
      | BoolLit l  -> (Bool, SBoolLit l)
      | Noexpr     -> (Void, SNoexpr)
      | Id s       -> (type_of_identifier s, SId s)
      | Assign(e1, e2) as ex -> 
          let (lt, e1' ) = expr e1 
          and (rt, e') = expr e2 in
          let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^ 
            string_of_typ rt ^ " in " ^ string_of_expr ex
          in (check_assign lt rt err, SAssign((lt, e1'), (rt, e')))
      | Unop(op, e) as ex -> 
          let (t, e') = expr e in
          let ty = match op with
            Neg when t = Int || t = Float -> t
          | Not when t = Bool -> Bool
          | _ -> raise (Failure ("illegal unary operator " ^ 
                                 string_of_uop op ^ string_of_typ t ^
                                 " in " ^ string_of_expr ex))
          in (ty, SUnop(op, (t, e')))
      | Binop(e1, op, e2) as e -> 
          let (t1, e1') = expr e1 
          and (t2, e2') = expr e2 in
          (* All binary operators require operands of the same type *)
          let same = t1 = t2 in
          (* Determine expression type based on operator and operand types *)
          let ty = match op with
            Add | Sub | Mult | Div when same && t1 = Int   -> Int
          | Add | Sub | Mult | Div when same && t1 = Float -> Float
		  | Add when same && t1 = String				   -> String
          | Equal | Neq            when same               -> Bool
          | Less | Leq | Greater | Geq
                     when same && (t1 = Int || t1 = Float) -> Bool
          | And | Or when same && t1 = Bool -> Bool
          | _ -> raise (
	          Failure ("illegal binary operator " ^
                       string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                       string_of_typ t2 ^ " in " ^ string_of_expr e))
          in (ty, SBinop((t1, e1'), op, (t2, e2')))

      (* maybe we can merge the FuncCall and GofuncCall tgt -> TODO *)
      | FuncCall(fname, args) as fcall -> 
          let fd = find_func fname in
          (* checking on if regular function *)
          let _ = if fd.isGo 
                  then raise (Failure ("regular call on gofunction " ^ fd.fname)) in
          let param_length = List.length fd.formals in
          if List.length args != param_length then
            raise (Failure ("expecting " ^ string_of_int param_length ^ 
                            " arguments in " ^ string_of_expr fcall))
          else let check_call (ft, _) e = 
            let (et, e') = expr e in 
            let err = "illegal argument found " ^ string_of_typ et ^
              " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
            in (check_assign ft et err, e')
          in 
          let args' = List.map2 check_call fd.formals args
          in (fd.typ, SFuncCall(fname, args'))
      | GofuncCall(fname, args) as gocall -> 
        let fd = find_func fname in
        (* checking on if gofunction *)
        let _ = if not fd.isGo 
                then raise (Failure ("go call on regular function " ^ fd.fname)) in
        let param_length = List.length fd.formals in
        if List.length args != param_length then
          raise (Failure ("expecting " ^ string_of_int param_length ^ 
                          " arguments in " ^ string_of_expr gocall))
        else let check_call (ft, _) e = 
          let (et, e') = expr e in 
          let err = "illegal argument found " ^ string_of_typ et ^
            " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
          in (check_assign ft et err, e')
        in
		if List.length args > 3 then raise(Failure ("no more than three args in gofunctioncall")) else 
        let args' = List.map2 check_call fd.formals args
        in (fd.typ, SGofuncCall(fname, args'))
      | Dot(e1,s) ->
        let (t',e1') = expr e1 in
        let binds = match t' with Struct(s) -> StringMap.find s struct_defs
            | _ -> raise (Failure "Not dot accessing struct type") in
        let rec find l name = match l with [] -> raise (Failure "field not found")
                      | (t,n)::tl -> if n = name then t
                      else find tl name in
        let ty = find binds s in (ty,SDot((t',e1'),s))
      | NArray(ty, e) ->
        let (t_e, s_e) = expr e in
        if t_e <> Int then raise (Failure "array size must be of int type")
        else (Array(ty), SNArray(ty, (t_e, s_e)))
      | ArrAccess(e1, e2) -> 
        let (t1, se1) = expr e1
        and (t2, se2) = expr e2 in
        let isArr = function 
          Array t -> t
          | _ -> raise (Failure "only array could be accessed with []")
        and isInt = function 
           Int -> Int
          | _ -> raise (Failure "array index must be of int type") in
        (isArr t1, SArrAccess((t1, se1), (isInt t2, se2)))
      | NChan(ty, e) ->
        let (t_e, s_e ) = expr e in
        let isInt = function
           Int -> Int
          | _ -> raise (Failure "length of channel needs to be of type int") in
        (Channel(ty), SNChan(ty, (isInt t_e, s_e))) 
      | Queue(e1, e2) -> 
        let (t1, se1) = expr e1
        and (t2, se2) = expr e2 in
        let same_t ft1 ft2 = if (ft1 = ft2 || ft2 = Void) then ft1
            else raise (Failure ("channel type: " ^ (string_of_typ ft1) ^
                          " and data type: " ^ (string_of_typ ft2) ^ " mismatch")) in
        let sexp = match (t1, t2) with
           (Channel sty, _) -> ((same_t sty t2), SDequeue((t1, se1), (t2, se2)))
          |(_, Channel ssty) -> ((same_t t1 ssty ), SEnqueue((t2, se2), (t1, se1)))
          | _ -> raise (Failure "-> can only be used on channels") in sexp
      | NStruct(t,el) -> 
        let n = match t with
            Struct(s) -> s
            | _ -> raise (Failure "struct type argument in initializing expected")
        in     
        let blist = if StringMap.mem n struct_defs then StringMap.find n struct_defs
              else raise (Failure "-> initalizing undefined struct") in
        let blist = List.map (fun (t,_) ->t) blist in
        let field_len = List.length blist in
        if List.length el != field_len then
                raise (Failure ("expecting " ^ string_of_int field_len  ^ 
                        " fields in intializing struct type " ^ n))
        else let check_field lft e = 
                let (et, e') = expr e in 
                let err = "fields of struct does not match " ^ string_of_typ lft ^
                  " expected " ^ string_of_typ et ^ " in " ^ string_of_expr e
                in (check_assign lft et err, e')
                in let fields' = List.map2 check_field blist el
        in (t, SNStruct(t, fields'))
	in

    let check_bool_expr e = 
      let (t', e') = expr e
      and err = "expected Boolean expression in " ^ string_of_expr e
      in if t' != Bool then raise (Failure err) else (t', e') 
    in

    (* Return a semantically-checked statement i.e. containing sexprs *)
    let rec check_stmt = function
        Expr e -> SExpr (expr e)
      | If(p, b1, b2) -> SIf(check_bool_expr p, check_stmt b1, check_stmt b2)
      | For(e1, e2, e3, st) ->
	  SFor(expr e1, check_bool_expr e2, expr e3, check_stmt st)
      | While(p, s) -> SWhile(check_bool_expr p, check_stmt s)
      | Return e -> let (t, e') = expr e in
        if t = func.typ then SReturn (t, e') 
        else raise (
	  Failure ("return gives " ^ string_of_typ t ^ " expected " ^
		   string_of_typ func.typ ^ " in " ^ string_of_expr e))
	    
	    (* A block is correct if each statement is correct and nothing
	       follows any Return statement.  Nested blocks are flattened. *)
      | Block sl -> 
          let rec check_stmt_list = function
              [Return _ as s] -> [check_stmt s]
            | Return _ :: _   -> raise (Failure "nothing may follow a return")
            | Block sl :: ss  -> check_stmt_list (sl @ ss) (* Flatten blocks *)
            | s :: ss         -> check_stmt s :: check_stmt_list ss
            | []              -> []
          in SBlock(check_stmt_list sl)

    in (* body of check_function *)
    { sisGo = func.isGo;
      styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      slocals  = func.locals;
      sbody = match check_stmt (Block func.body) with
	SBlock(sl) -> sl
      | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }
  in (sdefs,(globals, List.map check_function functions))
